from typing import List, Union

from zeno.classes.base import CamelModel


class PointsColors(CamelModel):
    color: List[int]
    domain: List[Union[int, float, str]]
    # "nominal" or "continuous" or "boolean"
    data_type: str


class Points2D(CamelModel):
    x: List[float]
    y: List[float]
    color: List[int]
    opacity: List[int]
    ids: Union[List[int], List[str]]
    domain: List[Union[int, float, str]]
    # "nominal" or "continuous" or "boolean"
    data_type: str



import secrets
from typing import List

import numpy as np
from sliceline.slicefinder import Slicefinder

from zeno.classes.slice import FilterPredicate, FilterPredicateGroup, Slice
from zeno.classes.slice_finder import SliceFinderRequest, SliceFinderReturn


def slice_finder(df, req: SliceFinderRequest):
    """Return slices of data with high or low metric values.

    Args:
        df (DataFrame): Zeno DataFrame with all metadata.
        req (SliceFinderRequest): Request with columns, metrics, and options.

    Returns a SliceFinderMetricReturn Object.
    """

    updated_df = df[
        list(set([str(col) for col in req.columns] + [str(req.metric_column)]))
    ].dropna()

    # Invert metric column if ascending.
    normalized_metric_col = np.array(updated_df[str(req.metric_column)], dtype=float)
    metric_max = np.max(normalized_metric_col)
    if req.order_by == "ascending":
        normalized_metric_col = metric_max - normalized_metric_col

    slice_finder = Slicefinder(alpha=req.alpha, k=20, min_sup=req.minimum_supp)
    slice_finder.fit(
        updated_df[[str(col) for col in req.columns]].to_numpy(), normalized_metric_col
    )

    if slice_finder.top_slices_ is None or slice_finder.top_slices_statistics_ is None:
        return SliceFinderReturn(slices=[], metrics=[], sizes=[], overall_metric=0)

    discovered_slices: List[Slice] = []
    slice_metrics: List[float] = []
    slice_sizes: List[int] = []

    for sli_i, sli in enumerate(slice_finder.top_slices_):
        # Rescale back to original metric.
        if req.order_by == "ascending":
            slice_metrics.append(
                abs(
                    slice_finder.top_slices_statistics_[sli_i]["slice_average_error"]
                    - metric_max
                )
            )
        else:
            slice_metrics.append(
                slice_finder.top_slices_statistics_[sli_i]["slice_average_error"]
            )

        slice_sizes.append(slice_finder.top_slices_statistics_[sli_i]["slice_size"])

        predicate_list = []
        for pred_i, sli_predicate in enumerate(sli):
            if sli_predicate is not None:
                join_val = "" if len(predicate_list) == 0 else "&"
                predicate_list.append(
                    FilterPredicate(
                        column=req.columns[pred_i],
                        operation="==",
                        value=sli_predicate,
                        join=join_val,
                    )
                )

        discovered_slices.append(
            Slice(
                slice_name="Generated Slice " + secrets.token_hex(nbytes=4),
                folder="",
                filter_predicates=FilterPredicateGroup(
                    predicates=predicate_list, join=""
                ),
            )
        )

    return SliceFinderReturn(
        slices=discovered_slices,
        metrics=slice_metrics,
        sizes=slice_sizes,
        overall_metric=slice_finder.average_error_
        if slice_finder.average_error_
        else 0,
    )



"""External API for Zeno. Includes decorator functions, return types, and options."""

import functools
from typing import Any, Callable, Dict, List, Union

from numpy.typing import NDArray
from pandas import DataFrame, Series
from pydantic import BaseModel


class ZenoOptions(BaseModel):
    """Parameters passed to Zeno test functions.

    Args:
        id_column (str): Column in dataframe with unique identifiers.
        data_column (str): Column in dataframe with either raw data or path to data.
        label_column (str): Column in dataframe with
                            either raw labels or path to labels.
        output_column (str): Column in dataframe with a given
                             model's raw output or path to output
        data_path (str): Path to directory with data files.
        label_path (str): Path to directory with label files.
        output_path (str): Path to directory with a given model's output.
        distill_columns (map[str, str]): Map from distill function name to
        distill column.
    """

    id_column: str
    data_column: str
    label_column: str
    output_column: str
    distill_columns: Dict[str, str]
    data_path: str
    label_path: str
    output_path: str


class ZenoParameters(BaseModel):
    """Options passed to the backend processing pipeline."""

    metadata: Union[DataFrame, str]
    # If run from command line, config_file will be a path to a config file.
    config_file: str = ""
    functions: Union[List[Callable], str] = []
    view: str = ""
    models: List[str] = []
    id_column: str = ""
    data_column: str = ""
    label_column: str = ""
    data_path: str = ""
    label_path: str = ""
    batch_size: int = 1
    cache_path: str = ""
    calculate_histogram_metrics = True
    editable: bool = True
    multiprocessing: bool = True
    serve: bool = True
    samples: int = 30
    port: int = 8000
    host: str = "localhost"

    class Config:
        arbitrary_types_allowed = True


class ModelReturn(BaseModel):
    """Return type for model functions.

    Args:
        model_output (Series | List): Model output for each sample.
        embedding (Series | List[List[float]] | List[NDArray] | NDArray | None):
        High-dimensional embedding for each sample. Optional.
        other_returns (Dict[str, Series | List] | None): Other returns from the model
        to be shown as metadata columns in the UI.
    """

    model_output: Union[Series, List[Any]]
    embedding: Union[Series, List[List[float]], List[NDArray], NDArray, None] = None
    other_returns: Union[Dict[str, Union[Series, List[Any]]], None] = None

    class Config:
        arbitrary_types_allowed = True


class DistillReturn(BaseModel):
    """Return type for distill functions

    Args:
        distill_output (Series | List): Distill outputs for each sample.
    """

    distill_output: Union[Series, List[Any]]

    class Config:
        arbitrary_types_allowed = True


class MetricReturn(BaseModel):
    """Return type for metric functions.

    Args:
        metric (float): Average metric over subset of data
    """

    metric: float

    class Config:
        arbitrary_types_allowed = True


def model(func: Callable[[str], Callable[[DataFrame, ZenoOptions], ModelReturn]]):
    """Decorator function for model functions.

    Args:
        func (Callable[[str], Callable[[DataFrame, ZenoOptions], ModelReturn]]):
        A function that that takes a model name and returns a model function, which
        itself returns a function that takes a DataFrame and ZenoOptions and returns a
        ModelReturn.
    """

    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    _wrapper.predict_function = True  # type: ignore
    return _wrapper


def distill(func: Callable[[DataFrame, ZenoOptions], DistillReturn]):
    """Deocrator function for distill functions.

    Args:
        func (Callable[[DataFrame, ZenoOptions], DistillReturn]):
        A function that takes a DataFrame and ZenoOptions and returns a DistillReturn.
    """

    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    _wrapper.distill_function = True  # type: ignore
    return _wrapper


def metric(func: Callable[[DataFrame, ZenoOptions], MetricReturn]):
    """Decorator function for metric functions.

    Args:
        func (Callable[[DataFrame, ZenoOptions], MetricReturn]):
            A metric function that takes a DataFrame and ZenoOptions and
            returns a MetricReturn with an average metric value and optional error rate.
    """

    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    _wrapper.metric_function = True  # type: ignore
    return _wrapper



"""Internal classes for Zeno."""

from typing import List, Optional, Tuple, Union

from pydantic import BaseModel

from zeno.classes.base import CamelModel, ZenoColumn
from zeno.classes.slice import FilterIds, FilterPredicateGroup, Slice


class ZenoSettings(CamelModel):
    view: str
    id_column: ZenoColumn
    label_column: ZenoColumn
    data_column: ZenoColumn
    samples: int
    calculate_histogram_metrics: bool
    total_size: int


class ZenoVariables(CamelModel):
    metrics: List[str]
    models: List[str]
    folders: List[str]


class StatusResponse(CamelModel):
    status: str
    done_processing: bool
    complete_columns: List[ZenoColumn]


class MetricKey(CamelModel):
    sli: Slice
    model: str
    metric: str


class MetricRequest(CamelModel):
    metric_keys: List[MetricKey]
    tag_ids: Optional[FilterIds] = None
    filter_ids: Optional[FilterIds] = None
    tag_list: Optional[List[str]] = None


class TableRequest(CamelModel):
    columns: List[ZenoColumn]
    slice_range: List[int]
    filter_predicates: FilterPredicateGroup
    sort: Tuple[Union[ZenoColumn, None], bool]
    tag_ids: FilterIds
    filter_ids: Optional[FilterIds] = None
    tag_list: List[str]


class PlotRequest(CamelModel):
    filter_predicates: FilterPredicateGroup
    tag_ids: FilterIds


class EmbedProject2DRequest(CamelModel):
    model: str
    column: ZenoColumn


class ColorsProjectRequest(CamelModel):
    column: ZenoColumn


class EntryRequest(BaseModel):
    id: Union[int, str]
    columns: List[ZenoColumn] = []



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zeno', 'zeno.classes', 'zeno.processing']

package_data = \
{'': ['*'],
 'zeno': ['frontend/*', 'frontend/build/*', 'frontend/build/assets/*']}

install_requires = \
['fastapi>=0.95.1,<0.96.0',
 'inquirer>=3.1.2,<4.0.0',
 'nest-asyncio>=1.5.6,<2.0.0',
 'opentsne>=0.7.1,<0.8.0',
 'pandas>=2.0.0,<3.0.0',
 'pathos>=0.3.0,<0.4.0',
 'requests>=2.28.1,<3.0.0',
 'setuptools',
 'tomli>=2.0.1,<3.0.0',
 'tqdm>=4.64.0,<5.0.0',
 'uvicorn>=0.22.0,<0.23.0',
 'websockets>=11.0,<12.0',
 'zeno-sliceline>=0.0.1,<0.0.2']

entry_points = \
{'console_scripts': ['zeno = zeno.runner:command_line']}

setup_kwargs = {
    'name': 'zenoml',
    'version': '0.6.0',
    'description': 'Interactive Evaluation Framework for Machine Learning',
    'long_description': '<img src="https://zenoml.com/img/zeno.png" width="250px"/>\n\n[![PyPI version](https://badge.fury.io/py/zenoml.svg)](https://badge.fury.io/py/zenoml)\n![Github Actions CI tests](https://github.com/zeno-ml/zeno/actions/workflows/ci.yml/badge.svg)\n[![MIT license](https://img.shields.io/badge/License-MIT-blue.svg)](https://lbesson.mit-license.org/)\n[![DOI](https://img.shields.io/badge/doi-10.1145%2F3544548.3581268-red)](https://cabreraalex.com/paper/zeno)\n[![Discord](https://img.shields.io/discord/1086004954872950834)](https://discord.gg/km62pDKAkE)\n\nZeno is a general-purpose framework for evaluating machine learning models.\nIt combines a **Python API** with an **interactive UI** to allow users to discover, explore, and analyze the performance of their models across diverse use cases.\nZeno can be used for any data type or task with [modular views](https://zenoml.com/docs/views/) for everything from object detection to audio transcription.\n\n### Demos\n\n|                                    **Image Classification**                                     |                                         **Audio Transcription**                                          |                                       **Image Generation**                                       |                                        **Dataset Chatbot**                                        |                                       **Sensor Classification**                                        |\n| :---------------------------------------------------------------------------------------------: | :------------------------------------------------------------------------------------------------------: | :----------------------------------------------------------------------------------------------: | :-----------------------------------------------------------------------------------------------: | :----------------------------------------------------------------------------------------------------: |\n|                                           Imagenette                                            |                                          Speech Accent Archive                                           |                                           DiffusionDB                                            |                                        LangChain + Notion                                         |                                              MotionSense                                               |\n| [![Try with Zeno](https://zenoml.com/img/zeno-badge.svg)](https://zeno-ml-imagenette.hf.space/) | [![Try with Zeno](https://zenoml.com/img/zeno-badge.svg)](https://zeno-ml-audio-transcription.hf.space/) | [![Try with Zeno](https://zenoml.com/img/zeno-badge.svg)](https://zeno-ml-diffusiondb.hf.space/) | [![Try with Zeno](https://zenoml.com/img/zeno-badge.svg)](https://zeno-ml-langchain-qa.hf.space/) | [![Try with Zeno](https://zenoml.com/img/zeno-badge.svg)](https://zeno-ml-imu-classification.hf.space) |\n|               [code](https://huggingface.co/spaces/zeno-ml/imagenette/tree/main)                |               [code](https://huggingface.co/spaces/zeno-ml/audio-transcription/tree/main)                |               [code](https://huggingface.co/spaces/zeno-ml/diffusiondb/tree/main)                |            [code](https://huggingface.co/spaces/zeno-ml/audio-transcription/tree/main)            |               [code](https://huggingface.co/spaces/zeno-ml/imu-classification/tree/main)               |\n\n<br />\n\nhttps://user-images.githubusercontent.com/4563691/220689691-1ad7c184-02db-4615-b5ac-f52b8d5b8ea3.mp4\n\n## Quickstart\n\nInstall the Zeno Python package from PyPI:\n\n```bash\npip install zenoml\n```\n\n### Command Line\n\nTo get started, run the following command to initialize a Zeno project. It will walk you through creating the `zeno.toml` configuration file:\n\n```bash\nzeno init\n```\n\nTake a look at the [configuration documentation](https://zenoml.com/docs/configuration) for additional `toml` file options like adding model functions.\n\nStart Zeno with `zeno zeno.toml`.\n\n### Jupyter Notebook\n\nYou can also run Zeno directly from Jupyter notebooks or lab. The `zeno` command takes a dictionary of configuration options as input. See [the docs](https://zenoml.com/docs/configuration) for a full list of options. In this example we pass the minimum options for exploring a non-tabular dataset:\n\n```python\nimport pandas as pd\nfrom zeno import zeno\n\ndf = pd.read_csv("/path/to/metadata/file.csv")\n\nzeno({\n    "metadata": df, # Pandas DataFrame with a row for each instance\n    "view": "audio-transcription", # The type of view for this data/task\n    "data_path": "/path/to/raw/data/", # The folder with raw data (images, audio, etc.)\n    "data_column": "id" # The column in the metadata file that contains the relative paths of files in data_path\n})\n\n```\n\nYou can pass a list of decorated function references directly Zeno as you add models and metrics.\n\n## Citation\n\nPlease reference our [CHI\'23 paper](https://arxiv.org/pdf/2302.04732.pdf)\n\n```bibtex\n@inproceedings{cabrera23zeno,\n  author = {Cabrera, Ángel Alexander and Fu, Erica and Bertucci, Donald and Holstein, Kenneth and Talwalkar, Ameet and Hong, Jason I. and Perer, Adam},\n  title = {Zeno: An Interactive Framework for Behavioral Evaluation of Machine Learning},\n  year = {2023},\n  isbn = {978-1-4503-9421-5/23/04},\n  publisher = {Association for Computing Machinery},\n  address = {New York, NY, USA},\n  url = {https://doi.org/10.1145/3544548.3581268},\n  doi = {10.1145/3544548.3581268},\n  booktitle = {CHI Conference on Human Factors in Computing Systems},\n  location = {Hamburg, Germany},\n  series = {CHI \'23}\n}\n```\n\n## Community\n\nChat with us on our [Discord channel](https://discord.gg/km62pDKAkE) or leave an issue on this repository if you run into any issues or have a request!\n',
    'author': 'Ángel Alexander Cabrera',
    'author_email': 'alex.cabrera@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://zenoml.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import datetime
import os
import pickle
import shutil
import sys
from contextlib import contextmanager
from importlib import util
from inspect import getmembers, isfunction
from pathlib import Path
from typing import Any, Callable, Dict, List, Union

import pandas as pd
import requests
import tomli

from zeno.api import ZenoParameters
from zeno.classes.base import MetadataType

VIEW_MAP_URL: str = "https://raw.githubusercontent.com/zeno-ml/instance-views/0.3/"
VIEWS_MAP_JSON: str = "views.json"


def read_pickle(file_name: str, cache_path: str, default):
    try:
        with open(os.path.join(cache_path, file_name), "rb") as f:
            return pickle.load(f)
    except FileNotFoundError:
        return default


def get_metadata_type(col: pd.Series) -> MetadataType:
    try:
        datetime.datetime.fromisoformat(str(col[0]))
        return MetadataType.DATETIME
    except ValueError:
        pass

    if col.dtype == "bool" or col.dtype == "boolean":
        return MetadataType.BOOLEAN

    try:
        unique = col.unique().tolist()
    except TypeError:
        return MetadataType.OTHER

    if len(unique) < 21:
        return MetadataType.NOMINAL

    if col.dtype in [
        "int64",
        "int32",
        "int16",
        "int8",
        "Int64",
        "Int32",
        "Int16",
        "Int8",
        "uint64",
        "uint32",
        "uint16",
        "uint8",
        "Uint64",
        "Uint32",
        "Uint16",
        "Uint8",
        "float64",
        "float32",
        "float16",
        "float8",
        "Float64",
        "Float32",
        "Float16",
        "Float8",
    ]:
        return MetadataType.CONTINUOUS

    return MetadataType.OTHER


def load_series(df, col_name, save_path):
    try:
        series = pd.read_pickle(save_path)
        col_name.metadata_type = get_metadata_type(series)
        df.loc[:, str(col_name)] = series
    except FileNotFoundError:
        df.loc[:, str(col_name)] = pd.Series([pd.NA] * df.shape[0], index=df.index)
    except EOFError:
        df.loc[:, str(col_name)] = pd.Series([pd.NA] * df.shape[0], index=df.index)


@contextmanager
def add_to_path(p):
    old_path = sys.path
    sys.path = sys.path[:]
    sys.path.insert(0, p)
    try:
        yield
    finally:
        sys.path = old_path


def read_config(args: Union[str, Dict, ZenoParameters]) -> ZenoParameters:
    params: ZenoParameters

    if isinstance(args, str):
        path = os.path.abspath(args)
        try:
            with open(path, "rb") as f:
                args_dict = tomli.load(f)
        except Exception:
            print("ERROR: Failed to read TOML configuration file.")
            sys.exit(1)

        params = ZenoParameters.parse_obj(args_dict)
        params.config_file = path

        # Change working directory to the directory of the config file.
        os.chdir(os.path.dirname(path))

    elif isinstance(args, dict):
        params = ZenoParameters.parse_obj(args)
    elif isinstance(args, ZenoParameters):
        params = args
    else:
        sys.exit(1)

    if params.cache_path == "":
        params.cache_path = "./.zeno_cache/"
    else:
        params.cache_path = params.cache_path
    os.makedirs(params.cache_path, exist_ok=True)

    # Try to get view from GitHub List, if not try to read from path and copy it.
    if params.view != "":
        view_dest_path = Path(os.path.join(params.cache_path, "view.mjs"))
        view_path = Path(params.view)
        if view_path.is_file():
            if view_dest_path.is_file():
                os.remove(view_dest_path)
            shutil.copyfile(view_path, view_dest_path)
        else:
            try:
                views_res = requests.get(VIEW_MAP_URL + VIEWS_MAP_JSON)
                views = views_res.json()
                url = VIEW_MAP_URL + views[params.view]
                with open(view_dest_path, "wb") as out_file:
                    content = requests.get(url, stream=True).content
                    out_file.write(content)
            except KeyError:
                print(
                    "ERROR: View not found in list or relative path."
                    " See available views at ",
                    "https://github.com/zeno-ml/instance-views/blob/main/views.json",
                )
                sys.exit(1)

    if params.id_column == "":
        print(
            "WARNING: no id_column specified, using index as id_column. If you are",
            "using a data_column, suggest using it as id_column.",
        )

    return params


def read_metadata(meta: Union[str, pd.DataFrame]) -> pd.DataFrame:
    """Return DataFrame or try to read it from file"""
    if isinstance(meta, pd.DataFrame):
        return meta
    elif isinstance(meta, str):
        meta_path = Path(os.path.realpath(meta))

        if meta_path.suffix == ".csv":
            return pd.read_csv(meta_path)
        elif meta_path.suffix == ".tsv":
            return pd.read_csv(
                meta_path, sep="\t", header=0, quoting=3, keep_default_na=False
            )
        elif meta_path.suffix == ".parquet":
            return pd.read_parquet(meta_path)
        elif meta_path.suffix == ".jsonl":
            return pd.read_json(meta_path, lines=True)

    print(
        "ERROR: Failed to read metadata file "
        + meta
        + "\n Should be one of .csv, .jsonl, or .parquet"
    )
    sys.exit(1)


def load_function(test_file: Path) -> List[Callable[..., Any]]:
    # To allow relative imports in test files,
    # add their directory to path temporarily.
    with add_to_path(os.path.dirname(os.path.abspath(test_file))):
        spec = util.spec_from_file_location(str(test_file), test_file)
        test_module = util.module_from_spec(spec)  # type: ignore
        spec.loader.exec_module(test_module)  # type: ignore

    functions: List[Callable[..., Any]] = []
    for _, func in getmembers(test_module):
        if isfunction(func):
            if (
                hasattr(func, "predict_function")
                or hasattr(func, "distill_function")
                or hasattr(func, "metric_function")
            ):
                functions.append(func)
    return functions


def read_functions(fns: Union[List[Callable], str]) -> List[Callable]:
    if isinstance(fns, list):
        return fns
    elif isinstance(fns, str):
        fn_path = Path(os.path.realpath(fns))
        if os.path.isfile(fn_path):
            return load_function(fn_path)
        elif os.path.exists(fn_path):
            # Add directory with tests to path for relative imports.
            fns = []
            for f in list(fn_path.rglob("*.py")):
                fns = fns + load_function(f)
            return fns
    return []


def is_notebook() -> bool:
    try:
        from IPython import get_ipython

        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except (NameError, ImportError):
        return False  # Probably standard Python interpreter



from typing import List

from zeno.classes.base import CamelModel, ZenoColumn
from zeno.classes.slice import Slice


class SliceFinderRequest(CamelModel):
    metric_column: ZenoColumn
    columns: List[ZenoColumn]
    order_by: str
    alpha: float
    minimum_supp: int


class SliceFinderReturn(CamelModel):
    slices: List[Slice]
    metrics: List[float]
    sizes: List[int]
    overall_metric: float
